/* $Workfile: GimleDemo.c $       $Revision: 14 $      $Modtime: 9/14/05 10:37a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    GIMLEDEMO.C

    A simple example console application demonstrating the use of OTX Driver
    with the Gimle-16-PCI-Plus board.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 16 E1/T1 receivers.
    The LI Internal HDLC controller is used to receive HDLC frames from the
    Line Interfaces.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_LOG_FILE "GimleDemo.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_TASK_REF 555
#define DEMO_SS7_FISU_FILTERING 1

#define DEMO_MAX_MULTIPLE_FRAME_READ 128

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OTX_INT32       nBoardType;
  int             boardNo;
  int             coLi;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL        bReceiveEqualizer;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_coHdlcRxFrames[32];
OTX_UINT32  g_coHdlcBadRxFrames[32];

struct {
  OTX_HANDLE hBoardController;
  OTX_HANDLE hQuadE1T1[4]; 
  OTX_HANDLE hTss0;
  OTX_HANDLE hLi[16];
} g_hPhysDevs;


struct {
  OTX_HANDLE hLiIn[16];
} g_hPipes;


struct {
  OTX_HANDLE hHdlcReceiver[16];
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************

  @func

  Dump the contents of the falc registers for the purpose of debugging..

*******************************************************************************/
void DemoDumpFalcRegisters(void)
{
  OTX_INT32 i;
  OTX_UINT32 nAddress, readVal;
  OTX_RESULT result;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    DemoLogf("\nFalc[%d] register dump\n=====================\n", i);
    for (nAddress = 0; nAddress < 0x100; nAddress++) {
      result = OtxDrvReadReg8(g_hPhysDevs.hLi[i], nAddress, &readVal);
      if (OTX_RESULT_CODE(result) != OTX_E_INVALIDARG) {
        DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }
      if ((nAddress & 0xf) == 0x0) {
        DemoLogf("0x%.2X  ", nAddress);
      }
      if ((nAddress & 0xf) == 0x8) {
        DemoLogf("  ");
      }
      DemoLogf(" %.2X", (OTX_UINT8) readVal);
      if ((nAddress & 0xf) == 0xf) {
        DemoLogf("\n");
      }
    }
    if ((nAddress & 0xf) == 0xf) {
      DemoLogf("\n");
    }
  }
}


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects  
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; 
  OTX_UINT32 nDeviceId;
  OTX_UINT32 nSubsystemId;

  // Connect to the Physical Board Controller
  DemoLogf("Opening Gimle-16-PCI-Plus board\n");
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController); DemoCheckRc(g_hPhysDevs.hBoardController, result);

  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss0); DemoCheckRc(g_hPhysDevs.hBoardController, result);
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvEnable(g_hPhysDevs.hTss0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hTss0, result);

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(g_hPhysDevs.hBoardController, result);
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_SUBSYSTEM_ID, &nSubsystemId); DemoCheckRc(g_hPhysDevs.hBoardController, result);
  DemoLogf("Board Device ID: %d, SubsystemId: 0x%.4X\n", nDeviceId, nSubsystemId);

  // The E1T1 devices are hosted by a quad E1T1 device. Open those four devices.
  for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
    DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hBoardController, result);
  }

  // Get handles to T1/E1 line interface devices 
  for (i = 0; i < g_demoEnv.coLi; i++) {
    nQuadE1T1 = i/4;
    DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_HANDLE hHostDevice;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create Hdlc Receivers logical devices hosted in the physical Line Interface devices
    hHostDevice = g_hPhysDevs.hLi[i];
    result = OtxDrvCreateLogicalDevice(hHostDevice, OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(hHostDevice, result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create an incoming Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[i]); DemoCheckRc(g_hPipes.hLiIn[i], result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      ts = 24; 
    } else {
      // E1 mode
      ts = 16; 
    }
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;

    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);

    // Connect HDLC Receiver to the external T1/E1 Pipes
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiIn[i], -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(g_hPipes.hLiIn[i], result);

  }

}




/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Initialize the Hdlc Receiver
    result = OtxHdlcReceiverInit(
               g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
               OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
               OTX_FALSE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
               DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);

  }

}


/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hBoardController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController, result);

  // Enable Quad E1T1 devices
  for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
    DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
    result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
#if 0 // Enable only Li1
    if (i == 1) {
      continue;
    }
#endif
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Enable Logical HDLC Devices 
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
  }
}

/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;   // AMI, B8ZS
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;   // AMI, B8ZS
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;    // F12 (D3/4 or SF), F4, ESF, F72 (SLC96)
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}




/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}


/*******************************************************************************
  @func

*******************************************************************************/
void  *DemoTimeStampTest(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  for (;;) {
    
    result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    //DemoLogf("%.8X%.8X (Current time)\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);
    
    // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
    #if defined(__linux__)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
    #elif defined(WIN32)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
    #endif
    
    // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
    nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
    
    // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
    nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
    
    nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
    ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
    
    if (ptCurrentDayAndTime) {
#if !defined(_WIN32_WCE)
      DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
        ptCurrentDayAndTime->tm_hour,
        ptCurrentDayAndTime->tm_min, 
        ptCurrentDayAndTime->tm_sec,
        nTimeStampMs.sUint32.nLow,
        asctime((struct tm *) ptCurrentDayAndTime)    // TODO: needs to be Osa function (OsaAscTime())
        );
#endif  // #if !defined(_WIN32_WCE)
    } else {
      DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
    }

    OsaThreadSleep(4000);
  }
  


  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
#if 0 // Single reads of events
      OtxEventDataS eventData;
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

#else // Multiple reads of events
#define DEMO_EVENT_DATA_BUF_SIZE 10
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

        assert(coEventsFetched <= coEventsFetched);

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }
#endif
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;
    
  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

#if 0 // Single Frame Read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
      OTX_UINT8  buf[1024];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nFrameSeq;
      OTX_UINT32 nFrameEndCode;
      OTX_DATETIME nTimeStamp;
      
      
      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
              
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId]++;
        }
        
        
        if (g_bHdlcVerbose) {
          DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");
        } else {
          if (nFrameEndCode) {
            
            g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
            
            DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            for (i = 0; i < nFrameLen && i < 256; i++) {
              DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
            
            if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
              DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
              otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
            }
            
            
          } else {
            if ((pFrame->cbOctets != (OTX_UINT32)g_cbHdlcFrame) && (pFrame->cbOctets != 3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            } else if (memcmp(pBuf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            }
          }
        } 
        
        // Read next frame (if any)
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);

      } 
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#else // Multiple frame read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
#define CB_BUF_SIZE 8192
      OTX_UINT8  buf[CB_BUF_SIZE];
      OTX_UINT8  *pBuf;
			OtxHdlcFrameInfoS *pFrame;
      OTX_UINT32 coFrames;
      OTX_UINT32 j;
      
      otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId] += coFrames;
        }

        if (g_bHdlcVerbose) {
          DemoLogf("%d HDLC Frame(s) Received, Sequence Number = %d\n", coFrames, pEventData->m_nParam);
        }
        
        pFrame = (OtxHdlcFrameInfoS *) buf;

        for (j = 0; j < coFrames; j++) {

					pBuf = pFrame->sqData;

          if (g_bHdlcVerbose) {
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
            for (i = 0; i < pFrame->cbOctets && i < CB_BUF_SIZE; i++) {
              DemoLogf("%.2X ", pBuf[i]);
            }
            DemoLogf("\n");
          } else {
            if (pFrame->nFrameEndCode) {
              
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
              for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                DemoLogf("%.2X ", pBuf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
              }
              
            }
          }  

          // Step pointer to next frame (if any)
					pFrame = OTX_STRUCT_ADD_BYTE_OFFSET(OtxHdlcFrameInfoS, pFrame, pFrame->nOffsetNext);
        }

        // Read next frame (if any)
        otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);

      } 
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadDataEx() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#endif

    
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

*******************************************************************************/
void DemoCloseDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;

  // Hdlc Receivers logical devices hosted in the physical Line Interface devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_hLogicDevs.hHdlcReceiver[i]) {
      result = OtxDrvClose(g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
    }
  }

  // Incoming Pipe representing the external T1/E1
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_hPipes.hLiIn[i]) {
      result = OtxDrvClose(g_hPipes.hLiIn[i]); DemoCheckRc(g_hPipes.hLiIn[i], result);
    }
  }

}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_GIMLE_16_PCI_PLUS;
  envP->coLi              = 16;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_T1;
  envP->bReceiveEqualizer = OTX_FALSE;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'H') {   // Longhaul mode
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /ME            : E1 Li mode\n");
  DemoLogf(" /MT            : T1 Li mode\n");
  DemoLogf(" /H             : Longhaul mode (default is shorthaul mode)\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentLi = 0;
  OTX_RESULT result;



  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" n/N Increment/decrement active Li (currently Li#%d selected)\n", nCurrentLi);
    DemoLogf(" e/E Enable/Disable current Li\n");
    DemoLogf(" c   Set Clock Source to recover from current Li\n");
    DemoLogf(" C   Set Clock Source to INTERNAL\n");
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" f   Force Resynchronization on both Lis\n");
    DemoLogf(" b   Reset Current HDLC Receiver\n");
    DemoLogf(" B   Reset Current E1T1 device\n");
    DemoLogf(" z   Set T1E1 Line Termination\n");
    DemoLogf(" Z   Toggle front end Amplifier ON or OFF for current Li\n");
    DemoLogf(" &   Read receive equalizer status\n");
    DemoLogf(" J   Set Receive Input Threshold for current Li\n");
    DemoLogf(" p   Set Loopback mode\n");
    DemoLogf(" g   Get current time\n");
    DemoLogf(" m/M Enable/Disable Monitor mode\n");
    DemoLogf(" F   Query line frequency of the current clock source\n");
    DemoLogf(" [   Read T1/E1 span voltage for current Li\n");
    break;

  case 'y':
    {

      DemoLogf("Change Li mode of all Lis \n");

      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        DemoLogf("Switching to E1 mode \n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_E1;
      } else {
        DemoLogf("Switching to T1 mode \n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_T1;
      }
      
      DemoCloseDevices();

      DemoConfigureE1T1Devices();
      
      DemoEnablePhysicalDevices();

      DemoCreateLogicalDevices();
      
      DemoConnectDevices();
      
      DemoInitializeLogicalDevices();
      
      DemoEnableLogicalDevices();
      
      DemoLogf("Done.\n");
    }
    break;

  case 'g':
    {
      OTX_DATETIME nTimeStamp;
      OTX_DATETIME nTimeStampMs;
      OTX_DATETIME nTimeStampSec;
      struct Osa_tm *ptCurrentDayAndTime;
      Osa_time_t nCurrentTimeSecs;
      
      result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      
      //DemoLogf("%.8X%.8X (Current time)\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);
      
      // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
      #if defined(__linux__)
        nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
      #elif defined(WIN32)
        nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
      #endif
      
      // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
      nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
      
      // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
      nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
      
      nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
      ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
      
      if (ptCurrentDayAndTime) {
#if !defined(_WIN32_WCE)
        DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
          ptCurrentDayAndTime->tm_hour,
          ptCurrentDayAndTime->tm_min, 
          ptCurrentDayAndTime->tm_sec,
          nTimeStampMs.sUint32.nLow,
          asctime((struct tm *) ptCurrentDayAndTime)    // TODO: needs to be Osa function (OsaAscTime())
          );
#endif  // #if !defined(_WIN32_WCE)
      } else {
        DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
      }
      

    }
    break;

  case 'G':
    {
      OsaThreadStart(DemoTimeStampTest);
    }
    break;


  case 'p':
    {
      OtxT1E1LoopE eLoopType;
      OTX_INT32 nLi;

      DemoLogf("Set Loopback mode for both Line Interfaces \n");

      DemoLogf("0: All Line Loops off                     \n");
      DemoLogf("1: Line Loop                              \n");
      DemoLogf("2: Remote Loop with Jitter Attenuation On \n");
      DemoLogf("3: Remote Loop with Jitter Attenuation Off\n");
      DemoLogf("4: Local Loop                             \n");
      
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      case 1:
        eLoopType = OTX_T1E1_LINE_LOOP;
        break;
      case 2:
        eLoopType = OTX_T1E1_REMOTE_ON_LOOP_JITTER_ATTEN;
        break;
      case 3:
        eLoopType = OTX_T1E1_REMOTE_OFF_LOOP_JITTER_ATTEN;
        break;
      case 4:
        eLoopType = OTX_T1E1_LOCAL_LOOP;
        break;
      default:
        DemoLogf("Invalid entry. Turning Loop off (OTX_T1E1_LOOPS_OFF)\n");
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      }

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Setting Loopback mode (%d) on Li #%d \n", (int)eLoopType, nLi);
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], eLoopType); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
      }

      DemoLogf("Done.\n");
    }
    break;

  case ':':
    {
      OTX_INT32 i;

      /*
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_E1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_E1;
      } else {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_T1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_T1;
      }
      */

      for (i = 0; i < g_demoEnv.coLi; i++) {
        
        // Set termination for the T1/E1 Front End
        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        
        // Configure Transceivers
        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        } else {
          // E1 mode
          result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        }
        result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        
      }

      DemoLogf("Done\n");
  
    }
    break;

  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'b':
    {
      DemoLogf("Reseting HdlcReceiver[%d]...\n",nCurrentLi);
      result = OtxDrvReset(g_hLogicDevs.hHdlcReceiver[nCurrentLi]); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 'B':
    {
      DemoLogf("Reseting Li #%d \n", nCurrentLi);
      result = OtxDrvReset(g_hPhysDevs.hLi[nCurrentLi]); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 'm':
    DemoLogf("Enabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'M':
    DemoLogf("Disabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'e':
    DemoLogf("Enable Li #%d \n", nCurrentLi);
    result = OtxDrvEnable(g_hPhysDevs.hLi[nCurrentLi], DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'E':
    DemoLogf("Disable Li #%d \n", nCurrentLi);
    result = OtxDrvDisable(g_hPhysDevs.hLi[nCurrentLi], DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Done.\n");
    }
    break;


  case 'F':
    {
      OTX_UINT32 nFrequencyCount;

      DemoLogf("Querying frequency of current clock source ...\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_FREQUENCY_COUNTER, &nFrequencyCount); DemoCheckRc(g_hPhysDevs.hBoardController , result);

      if (OTX_SUCCESS(result)) {
        OTX_UINT32 nFrequency;

        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          nFrequency = (nFrequencyCount+1)*193/2048;
        } else {
          // E1 mode
          nFrequency = (nFrequencyCount+1)/8;
        }
        DemoLogf("Line frequency is %d Hz (nFrequencyCount=%d)\n", nFrequency, nFrequencyCount);
      }

      DemoLogf("Done.\n");
    }
    break;


  case 'n':
    {
      // Select current nCurrentLi
      nCurrentLi++;
      if (nCurrentLi > (g_demoEnv.coLi-1)) {
        // Wrap 
        nCurrentLi = 0;
      }
      
      DemoLogf("Li#%d Selected \n", nCurrentLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentLi
      nCurrentLi--;
      if (nCurrentLi < 0) {
        // Wrap
        nCurrentLi = g_demoEnv.coLi-1;
      }
      
      DemoLogf("Li#%d Selected \n", nCurrentLi);
    }
    break;
    

  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'u':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in QuadE1T1%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hQuadE1T1[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nCurrentLi], result);

      DemoLogf("QuadE1T1%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'R':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevs.hBoardController, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'j':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in QuadE1T1%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hQuadE1T1[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nCurrentLi], result);

      DemoLogf("Wrote %.2X to QuadE1T1%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'W':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;

  case 'v':
    {
      OTX_INT32 i;
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
      for (i = 0; i < 32; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
          g_coHdlcRxFrames[i] = 0;
          g_coHdlcBadRxFrames[i] = 0;
        }
      }
    }
    break;

  case 'V':
    {
      OTX_INT32 i;
      for (i = 0; i < 32; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("Tot[%.2d]=%.6d (Bad[%.2d]=%.6d)  ", i, g_coHdlcRxFrames[i], i, g_coHdlcBadRxFrames[i]);
          if (i && i % 2) {
            DemoLogf("\n");
          }
        }
      }
      DemoLogf("\n");
    }
    break;

  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;

      DemoLogf("Setting Clock Source to recover from current Li (#%d)\n", nCurrentLi);

      eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + nCurrentLi ); 
      result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      DemoLogf("Done.\n");
    }
    break;
  
  case 'C':
    {
      OtxBrdClockSourceE  eClkSource;

      DemoLogf("Setting Clock Source to INTERNAL\n");

      eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_INTERNAL ); 
      result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      DemoLogf("Done.\n");
    }
    break;
  

  case 'O':
    {
      char szBuf[12];
      OTX_INT32 nCmd;
      
      DemoLogf("Driver Debug Command. Which Cmd?\n");
      DemoLogf("0: Dump QuadFalc register\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      DemoLogf("Issuing Driver Debug Command %d on QuadE1T1%d \n", nCmd, nCurrentLi);
      OtxDrvDbgDump(g_hPhysDevs.hQuadE1T1[nCurrentLi], nCmd);
    }
    break;




#if 1 // Just for debugging
  case '{':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      OTX_UINT32 nReadVal;

      DemoLogf("Read from TSS control register. Which control register? [0x%.2X] ", nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x01, nTssRegister); DemoCheckRc(g_hPhysDevs.hTss0, result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x02, 0x00); DemoCheckRc(g_hPhysDevs.hTss0, result);

      result = OtxDrvReadReg8(g_hPhysDevs.hTss0,  0x03, &nReadVal);  DemoCheckRc(g_hPhysDevs.hTss0, result);

      DemoLogf("Read 0x%.2X from TSS Control Register 0x%.2X\n", (OTX_UINT8)nReadVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;

  case '}':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      static OTX_UINT8 nWriteVal = 0;

      DemoLogf("Write to TSS control register. Which control register? [0x%.2X] ", nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      DemoLogf("Which value? [0x%.2X] ", nWriteVal);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nWriteVal = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x01, nTssRegister); DemoCheckRc(g_hPhysDevs.hTss0, result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x02, 0x00); DemoCheckRc(g_hPhysDevs.hTss0, result);

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0,  0x03, nWriteVal);  DemoCheckRc(g_hPhysDevs.hTss0, result);

      DemoLogf("Wrote 0x%.2X to TSS Control Register 0x%.2X\n", nWriteVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;

  case 'i':
    {
      DemoLogf("Enable TSS... \n");
      result = OtxDrvEnable(g_hPhysDevs.hTss0, OTX_TASK_SYNC,0); DemoCheckRc(g_hPhysDevs.hTss0, result);
      DemoLogf("Done.\n");
    }
    break;

  case 'I':
    {
      DemoLogf("Disable TSS... \n");
      result = OtxDrvDisable(g_hPhysDevs.hTss0, OTX_TASK_SYNC,0); DemoCheckRc(g_hPhysDevs.hTss0, result);
      DemoLogf("Done.\n");
    }
    break;

  case '_':
    {
      DemoLogf("Reseting TSS... \n");
      result = OtxDrvReset(g_hPhysDevs.hTss0); DemoCheckRc(g_hPhysDevs.hTss0, result);
      DemoLogf("Done.\n");
    }
    break;
#endif

  case '&':
    {
      OTX_UINT32 readVal;
      OTX_INT32 nLi;
      char *pszEqStatus;
      
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        result = OtxDrvReadReg8(g_hPhysDevs.hLi[nLi], 115, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        switch ((readVal & 0xC0) >> 6) {
        case 0x00:
          pszEqStatus = "not valid, still adapting";
          break;
        case 0x01:
          pszEqStatus = "valid";
          break;
        case 0x02:
          pszEqStatus = "not valid";
          break;
        case 0x03:
          pszEqStatus = "valid but high noise floor";
          break;
        default:
          pszEqStatus = "unknown";
          break;
        }
        DemoLogf("Receive Equalizer of Li#%d: status=%s attenuation=%2.1fdB (0x%.2X)\n", nLi, pszEqStatus, (readVal & 0x1F)*1.4, readVal );
      }
      
    }
    break; 
    
  case 'J':
    {
      char strBuf[12];
      OTX_UINT32 nThresholdInMilliVolts;

      DemoLogf("Set Receive Input Threshold for Li%d \n", nCurrentLi);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Threshold (%d mV): ", nThresholdInMilliVolts);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nThresholdInMilliVolts = strtoul(strBuf, NULL, 0);

        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        DemoLogf("Receive Input Threshold is changed to %d mV.\n", nThresholdInMilliVolts);
      } else {
        DemoLogf("The Receive Input Threshold remains unchanged.\n");
      }

    }
    break;

  case 'Z':
    {
      OTX_BOOL bAmpOn;
      
      result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, &bAmpOn); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      
      if (bAmpOn) {
        bAmpOn=OTX_FALSE;
      } else {
        bAmpOn=OTX_TRUE;
      }
      
      DemoLogf("Turning front end Amplifier %s for Li%d \n", bAmpOn ? "ON": "OFF", nCurrentLi);
      
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, bAmpOn); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      
      DemoLogf("Done.\n");
      
    }
    break;


  
  case 'd':   // dump falc registers
    DemoDumpFalcRegisters();
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;


    // E1/T1 span voltmeter
  case '[':
    {
      OTX_UINT32 nVoltage;
      
      DemoLogf("Reading T1/E1 Span Voltage for Li%d: \n", nCurrentLi);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_LINE_VOLTAGE, &nVoltage); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      
      DemoLogf("  Voltage = %d \n", nVoltage);
      
    }
    break;


/*
  // For testing E1/T1 span voltmeter
  case '[':
    {
      OTX_UINT32 readVal;
      static OTX_UINT32 nAddress = 0x006;
      static OTX_UINT32 nValue = 0;
      OTX_UINT32 i;
      OTX_UINT8 nSampleVal = 0;

      result = OtxDrvReadReg16(g_hPhysDevs.hBoardController, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      nValue = readVal & ~0x00CC;
      nValue |= 0x0080;  // Set CS# high
      result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      DemoLogf("Test Volt meter:\n");
      DemoLogf("0 = Clear\n");
      DemoLogf("1 = Measure\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        nValue |= 0x0040;  // Set bit 6 to discharge capacitor
        result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
        OsaThreadSleep(300);
        nValue &= ~0x0040;  // Clear bit 6 to allow charge
        result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
        //break;
      case 1:
        nValue &= ~0x0080;  // Set bit 7 low to activate CS#
        result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
        //OsaThreadSleep(1);

        for (i = 0; i < 8; i++) {

          // Bring Clock High
          nValue |= 0x0008;  // Set bit 3 to bring clock high
          result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
          //OsaThreadSleep(1);

          // Bring Clock Low
          nValue &= ~0x0008;  // Set bit 3 to bring clock low
          result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
          //OsaThreadSleep(1);
        }

        nValue |= 0x0080;  // Set bit 7 high to dectivate CS#
        result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
        //OsaThreadSleep(1);

#if 0
        // This needs take at least 17us
        for (i = 0; i < 40; i++) {

          // Bring Clock High
          nValue |= 0x0008;  // Set bit 3 to bring clock high
          result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
          OsaThreadSleep(1);
          
          // Bring Clock Low
          nValue &= ~0x0008;  // Set bit 3 to bring clock low
          result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
          OsaThreadSleep(1);

        }
#else
        OsaThreadSleep(1);
#endif

        nValue &= ~0x0080;  // Set bit 7 low to activate CS#
        result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
        //OsaThreadSleep(1);


        for (i = 0; i < 8; i++) {

          // Read Data bit (from FPGA Status register DONE_7)
          result = OtxDrvReadReg16(g_hPhysDevs.hBoardController, 0x001C, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController, result);
          //OsaThreadSleep(1);

          if (readVal & 0x0010) {
            nSampleVal |= (0x01 << (7-i));
          }

          // Bring Clock High
          nValue |= 0x0008;  // Set bit 3 to bring clock high
          result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
          //OsaThreadSleep(1);
          
          // Bring Clock Low
          nValue &= ~0x0008;  // Set bit 3 to bring clock low
          result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
          //OsaThreadSleep(1);

        }

        nValue |= 0x0080;  // Set bit 7 high to dectivate CS#
        result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
        //OsaThreadSleep(1);

        DemoLogf("Sample Value = %u\n", nSampleVal);

        break;
      default:
        DemoLogf("Invalid entry.\n");
        break;
      }
            
      DemoLogf("Done.\n");
    }
    break;
*/    
  
  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("GIMLEDEMO, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Gimle-16-PCI-Plus Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hBoardController, result);
  //result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hBoardController, result);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}
